# -*- coding:utf-8 -*-
"""Constant information about the sync."""

PACKAGE_NAME = 'rayvision_sync'

TASK_STATUS_DESCRIPTION = {
    "0": {
        "0": "等待中",
        "1": "Waiting"
    },
    "5": {
        "0": "渲染中",
        "1": "Rendering"
    },
    "8": {
        "0": "预处理中",
        "1": "Preprocessing"
    },
    "10": {
        "0": "停止",
        "1": "Stop"
    },
    "20": {
        "0": "欠费停止",
        "1": "Arrearage-stop"
    },
    "23": {
        "0": "超时停止",
        "1": "Timeout stop"
    },
    "25": {
        "0": "已完成",
        "1": "Done"
    },
    "30": {
        "0": "已完成(有失败帧)",
        "1": "Done(with failed frame)"
    },
    "35": {
        "0": "放弃",
        "1": "Abort"
    },
    "40": {
        "0": "等待全速渲染",
        "1": "Test done"
    },
    "45": {
        "0": "失败",
        "1": "Failed"
    }
}

TASK_END_STATUS_CODE_LIST = ['10', '20', '23', '25', '30', '35', '45']

TRANSFER_LOG = "RAYVISION_LOG"
RAYVISION_DB = "RAYVISION_DB"

WINDOWS_LOCAL_ENV = "USERPROFILE"
LINUX_LOCAL_ENV = "HOME"
RENDERFARM_SDK = "renderfarm_sdk"

ENGINE_TYPE = ['aspera', 'raysync']


"""Manage operations after generating tasks."""

from builtins import str
import logging

from rayvision_sync.utils import get_task_status_description

from rayvision_sync.constants import TASK_END_STATUS_CODE_LIST


# pylint: disable=useless-object-inheritance
class RayvisionManageTask(object):
    """Processing asset information for a task."""

    def __init__(self, query):
        """Instantiate API interface."""
        self._query = query
        self.logger = logging.getLogger(__name__)

    def is_task_end(self, task_id):
        """Check if the task rendering ends.

        Args:
            task_id (int): Task id

        Returns:
            bool: True: end of task rendering, False/None: Task rendering is
                not over.

        """
        result = None
        task_status_list = self.get_task_status([task_id])
        task_status_codes = self.find_task_status_codes(task_status_list)
        if task_status_codes:
            for task_status_code in task_status_codes:
                if task_status_code not in TASK_END_STATUS_CODE_LIST:
                    result = False
                    break
            if result is not False:
                result = True

        return result

    def get_task_status(self, task_id_list):
        """Get information about each task in the task id list.

        Call the API interface to get the ``items`` information of each task,
        and process it.

        Args:
            task_id_list (list of int): Task id list.

        Returns:
            list: Information about each task id.
                e.g.:
                    [
                        {
                            "task_id":"111",
                            "task_status_code":"25",
                            "task_status_text":"render_task_status_25",
                            "task_status_description":"Done",
                            "is_opener":"0",
                            "output_file_name":"111_test",
                            "sub_task_status":[]
                        },
                        {
                            "task_id":"222",
                            "task_status_code":"0",
                            "task_status_text":"render_task_status_0",
                            "task_status_description":"Waiting",
                            "is_opener":"1",
                            "output_file_name":None,
                            "sub_task_status":[]
                        },
                    ]

        """
        task_info_list = self._query.task_info(task_id_list).get(
            'items',
            [])
        task_status_list = self.task_info_iterater(task_info_list)
        return task_status_list

    def task_info_iterater(self, task_info_list):
        """Item information for each task, extracted and organized.

        Args:
            task_info_list (list): Some details about the task.
                e.g.:
                    [
                        {
                            "sceneName": "demo_scenc.mb",
                            "id': 6419169,
                            "taskAlias': "2W6419169",
                            "taskStatus': 0,
                            "statusText': "render_task_status_0",
                            "preTaskStatus": None,
                            "preStatusText": None,
                            "totalFrames": 10,
                            "abortFrames": 0,
                            "executingFrames": 0,
                            "doneFrames": 0,
                            "failedFrames": 0,
                            "framesRange": "1-10[1]",
                            "projectName": "Project1",
                            "renderConsume": None,
                            "taskArrears": 0.0,
                            "submitDate": 1563356906040,
                            "startTime": None,
                            "completedDate": None,
                            "renderDuration": 0,
                            "userName": "mxinye12",
                            "producer": ",
                            "taskLevel": 80,
                            "taskUserLevel": 0,
                            "taskLimit": 3,
                            "taskOverTime": 12,
                            "overTimeStop": 28800,
                            "userId": 100093088,
                            "outputFileName": "6419169_demo_scenc",
                            "munuTaskId": "2019071702241",
                            "layerParentId": 0,
                            "cgId": 2000,
                            "userAccountConsume": None,
                            "couponConsume": None,
                            "qyCouponConsume": None,
                            "isOpen": 0,
                            "taskType": "Render",
                            "renderCamera": "perspShape",
                            "cloneParentId": 0,
                            "cloneOriginalId": 0,
                            "shareMainCapital": 0,
                            "taskRam": 64,
                            "respRenderingTaskList": None,
                            "layerName": ",
                            "taskTypeText": "render_major_picture_task",
                            "locationOutput": ",
                            "isDelete": 1,
                            "channel": 4,
                            "remark": "gdgsgsg",
                            "isOverTime": 0,
                            "taskKeyValueVo": {
                                "tiles": None,
                                "allCamera": None,
                                "renderableCamera": None
                            },
                            "waitingCount": None
                        },
                        {}
                    ]

        Returns:
            list: Information about each task id.
                e.g.:
                    [
                        {
                            "task_id":"111",
                            "task_status_code":"25",
                            "task_status_text":"render_task_status_25",
                            "task_status_description":"Done",
                            "is_opener":"0",
                            "output_file_name":"111_test",
                            "sub_task_status":[]
                        },
                        {
                            "task_id":"222",
                            "task_status_code":"0",
                            "task_status_text":"render_task_status_0",
                            "task_status_description":"Waiting",
                            "is_opener":"1",
                            "output_file_name":"fasfafe",
                            "sub_task_status":[]
                        },
                    ]

        """
        task_status_list = []
        for task_info in task_info_list:
            task_status_dict = {}

            task_id = task_info.get('id')
            task_status_code = str(task_info.get('taskStatus'))  # e.g. 25.
            # e.g. "render_task_status_25".
            task_status_text = task_info.get('statusText')
            # 0: have not sub_task_status; 1:have sub_task_status.
            is_opener = task_info.get('isOpen')
            # Download directory name.
            output_file_name = task_info.get('outputFileName')
            task_status_description = (
                get_task_status_description(task_status_code))
            sub_task_status = []
            if int(is_opener) == 1:
                task_info_list_new = task_info.get('respRenderingTaskList',
                                                   [])
                sub_task_status = self.task_info_iterater(
                    task_info_list_new)

            task_status_dict['task_id'] = str(task_id)
            task_status_dict['task_status_code'] = str(task_status_code)
            task_status_dict['task_status_text'] = str(task_status_text)
            task_status_dict['task_status_description'] = (
                task_status_description)
            task_status_dict['is_opener'] = str(is_opener)
            task_status_dict['output_file_name'] = output_file_name
            task_status_dict['sub_task_status'] = sub_task_status

            task_status_list.append(task_status_dict)

        return task_status_list

    def output_file_names(self, task_status_list):
        """Get the name of the output scene to download.

        Args:
            task_status_list (list): Task information list.

        Returns:
            list: Output scene name.
                e.g.:
                    [
                        "block_scene",
                        "name_scene",
                    ]

        """
        output_file_names = []
        for task_status_dict in task_status_list:
            output_file_name = task_status_dict.get('output_file_name',
                                                    None)
            is_opener = task_status_dict.get('is_opener')
            sub_task_status = task_status_dict.get('sub_task_status', [])

            if int(is_opener) == 1:  # Have sub tasks.
                if sub_task_status:
                    output_file_list_sub = (
                        self.output_file_names(sub_task_status))
                    output_file_names.extend(output_file_list_sub)
            else:
                if output_file_name is not None:
                    output_file_names.append(output_file_name)

        return output_file_names

    def find_task_status_codes(self, task_status_list):
        """Get the task status code from the task information list.

        Args:
            task_status_list (list): Task information list.
                e.g.:
                    [
                        {
                            "task_id":"111",
                            "task_status_code":"25",
                            "task_status_text":"render_task_status_25",
                            "task_status_description":"Done",
                            "is_opener":"0",
                            "output_file_name":"111_test",
                            "sub_task_status":[]
                        },
                        {
                            "task_id":"222",
                            "task_status_code":"0",
                            "task_status_text":"render_task_status_0",
                            "task_status_description":"Waiting",
                            "is_opener":"1",
                            "output_file_name":None,
                            "sub_task_status":[]
                        },
                    ]

        Returns:
            list: Task status code list.
                e.g.:
                    [
                        "25",
                        "10",
                        "35"
                    ]

        """
        task_status_codes = []
        for task_status_dict in task_status_list:
            task_status_code = task_status_dict.get('task_status_code',
                                                    None)
            is_opener = task_status_dict.get('is_opener')
            sub_task_status = task_status_dict.get('sub_task_status', [])

            if int(is_opener) == 1:  # Have sub tasks.
                if sub_task_status:
                    task_status_code_sub = (
                        self.find_task_status_codes(sub_task_status))
                    task_status_codes.extend(task_status_code_sub)
            else:
                if task_status_code is not None:
                    task_status_codes.append(task_status_code)

        return task_status_codes



"""Upload models.

Upload the scene's configuration file and asset file.

"""

# Import built-in modules
import configparser
import os
import subprocess
from concurrent.futures import ThreadPoolExecutor
from threading import Thread

# Import local modules
from rayvision_sync import RayvisionTransfer
from rayvision_sync.constants import TRANSFER_LOG, RENDERFARM_SDK, WINDOWS_LOCAL_ENV, LINUX_LOCAL_ENV, RAYVISION_DB
from rayvision_sync.exception import RayvisionError, UnsupportedDatabaseError
from rayvision_sync.utils import create_transfer_params
from rayvision_sync.utils import read_ini_config
from rayvision_sync.utils import run_cmd
from rayvision_sync.utils import str2unicode
from rayvision_sync.utils import upload_retry


class RayvisionUpload(object):
    """Upload files.

    Upload configuration files and asset files.

    """

    def __init__(self, api, db_config_path=None):
        """Initialize instance."""
        params = create_transfer_params(api)
        self.api = api
        self.trans = RayvisionTransfer(**params)
        self.logger = self.trans.logger

        # load db config ini
        self.transfer_log_path, self.redis_config, self.sqlite_config, self.database_config = \
            self.load_db_config(db_config_path)

        custom_db_path = self.database_config.get("db_path")
        db_dir = self._check_and_mk(custom_db_path)
        self.check_transfer_log_path(self.transfer_log_path)
        self.db_ini = os.path.join(db_dir, 'db_ini')
        self._db = os.path.join(db_dir, 'db')

    def check_transfer_log_path(self, transfer_log_path):
        """Check the log location of the transport engine."""
        rayvision_log_path = os.environ.get(TRANSFER_LOG, "")
        if bool(transfer_log_path) and os.path.exists(transfer_log_path):
            transfer_path = transfer_log_path
            if rayvision_log_path != transfer_log_path:
                os.environ.update({TRANSFER_LOG: transfer_path})
                subprocess.Popen('setx %s "%s" /m' % (TRANSFER_LOG, transfer_path), shell=True)
        elif os.path.exists(rayvision_log_path):
            transfer_path = rayvision_log_path
        else:
            if self.api.user_info['local_os'] == "windows":
                transfer_path = os.path.join(os.environ[WINDOWS_LOCAL_ENV], RENDERFARM_SDK)
            else:
                transfer_path = os.path.join(os.environ[LINUX_LOCAL_ENV], RENDERFARM_SDK)

            os.environ.update({TRANSFER_LOG: transfer_path})
            subprocess.Popen('setx %s "%s" /m' % (TRANSFER_LOG, transfer_path), shell=True)

        return transfer_path

    def _check_and_mk(self, custom_db_path):
        """Check the path to the DB data file generated by the upload asset."""
        rayvision_db_env = os.environ.get(RAYVISION_DB, "")
        if bool(custom_db_path) and os.path.exists(custom_db_path):
            db_path = custom_db_path
        elif os.path.exists(rayvision_db_env):
            db_path = rayvision_db_env
        else:
            if self.api.user_info['local_os'] == "windows":
                db_path = os.path.join(os.environ[WINDOWS_LOCAL_ENV], RENDERFARM_SDK)
            else:
                db_path = os.path.join(os.environ[LINUX_LOCAL_ENV], RENDERFARM_SDK)

        return db_path

    def create_db_ini(self, upload_json_path):
        """Create the database configuration file.

        Args:
            upload_json_path (str): Upload json path.

        Returns:
            str: Configuration file path.

        """
        db_type = self.database_config.get("type", "sqlite").strip().lower()
        try:
            # Prevent multi-thread lock-free problems
            if not os.path.exists(self.db_ini):
                os.makedirs(self.db_ini)
        except:
            pass
        time_temp = os.path.split(os.path.dirname(upload_json_path))[-1]
        db_path = os.path.join(self._db, "%s.db" % time_temp)
        config_ini = configparser.ConfigParser()
        config_ini['database'] = {
            "on": self.database_config.get("on", "true"),
            "platform_id": self.trans.platform,
            "type": db_type
        }
        config_ini['redis'] = {
            "host": self.redis_config.get("host", "127.0.0.1"),
            "port": self.redis_config.get("port", 6379),
            "password": self.redis_config.get("password", ""),
            "table_index": self.redis_config.get("table_index", ""),
            "timeout": self.redis_config.get("timeout", 5000)
        }
        config_ini['sqlite'] = {
            "db_path": db_path,
            "temporary": self.sqlite_config.get("temporary", "false")
        }
        if db_type == "redis":
            db_ini_path = os.path.join(self.db_ini, "db_redis.ini")
        elif db_type == "sqlite":
            db_ini_path = os.path.join(self.db_ini, "%s.ini" % time_temp)
        else:
            error_data_msg = "{} is not a supported database, only support 'redis' or 'sqlite'".format(db_type)
            raise UnsupportedDatabaseError(error_data_msg)

        with open(db_ini_path, 'w') as configfile:
            config_ini.write(configfile)
        return db_ini_path

    def upload(self, task_id, task_json_path, tips_json_path, asset_json_path,
               upload_json_path, max_speed=None, transmit_type="upload_json",
               engine_type="aspera", server_ip=None, server_port=None):
        """Run the cmd command to upload the configuration file.

        Args:
            task_id (str, optional): Task id.
            task_json_path (str, optional): task.json file absolute path.
            tips_json_path (str, optional): tips.json file absolute path.
            asset_json_path (str, optional): asset.json file absolute path.
            upload_json_path (str, optional): upload.json file absolute path.
            max_speed (str): Maximum transmission speed, default value
                is 1048576 KB/S.
            transmit_type (str): transmit type:
                1. upload_json: upload from json file,in this type, next remote will not used.
                2. upload_list: upload from file list.
            engine_type (str, optional): set engine type, support "aspera" and "raysync", Default "aspera".
            server_ip (str, optional): transmit server host,
                if not set, it is obtained from the default transport profile.
            server_port (str, optional): transmit server port,
                if not set, it is obtained from the default transport profile.

        Returns:
            bool: True is success, False is failure.

        """
        config_file_list = [
            task_json_path,
            tips_json_path,
            asset_json_path,
            upload_json_path
        ]
        result_config = self.upload_config(task_id, config_file_list, max_speed,
                                           engine_type=engine_type, server_ip=server_ip, server_port=server_port)
        if not result_config:
            return False
        result_asset = self.upload_asset(upload_json_path, max_speed, transmit_type,
                                         engine_type=engine_type, server_ip=server_ip, server_port=server_port)
        if not result_asset:
            return False
        return True

    def upload_config(self, task_id, config_file_list, max_speed=None,
                      engine_type="aspera", server_ip=None, server_port=None):
        """Run the cmd command to upload configuration profiles.

        Args:
            task_id (str): Task id.
            config_file_list (list): Configuration file path list.
            max_speed (str): Maximum transmission speed, default value
                is 1048576 KB/S.
            engine_type (str, optional): set engine type, support "aspera" and "raysync", Default "aspera".
            server_ip (str, optional): transmit server host,
                if not set, it is obtained from the default transport profile.
            server_port (str, optional): transmit server port,
                if not set, it is obtained from the default transport profile.

        Returns:
            bool: True is success, False is failure.

        """
        transmit_type = "upload_path"
        max_speed = max_speed if max_speed is not None else "1048576"

        for config_path in config_file_list:
            local_path = str2unicode(config_path)

            config_basename = os.path.basename(config_path)
            server_path = '/{0}/cfg/{1}'.format(task_id, config_basename)
            server_path = str2unicode(server_path)

            if not os.path.exists(local_path):
                self.logger.info('%s is not exists.', local_path)
                continue
            cmd_params = [transmit_type, local_path, server_path, max_speed,
                          'false', 'config_bid']
            cmd = self.trans.create_cmd(cmd_params, engine_type=engine_type, server_ip=server_ip,
                                        server_port=server_port)

            times = 0
            while True:
                result = run_cmd(cmd, flag=True, logger=self.logger)
                if result:
                    if times == 9:
                        raise RayvisionError(20004, "%s upload failed" %
                                             config_path)
                    times += 1
                else:
                    break
        return True

    @upload_retry
    def upload_asset(self, upload_json_path, max_speed=None, is_db=True,
                     engine_type="aspera", server_ip=None, server_port=None,
                     transmit_type="upload_json"):
        """Run the cmd command to upload asset files.

        Args:
            upload_json_path (str): Path to the upload.json file.
            max_speed (str): Maximum transmission speed, default value
                is 1048576 KB/S.
            is_db (bool): Whether to produce local database record upload file.
            engine_type (str): Transport engine type, supports "aspera" and "raysync".
            server_ip (str): Transport server IP.
            server_port (str): Transport server port.
            transmit_type (str): transmit type:
                1. upload_json: upload from json file,in this type, next remote will not used.
                2. upload_list: upload from file list.

        Returns:
            bool: True is success, False is failure.

        """
        max_speed = max_speed if max_speed is not None else "1048576"
        cmd_params = [transmit_type, upload_json_path, '/', max_speed,
                      'false', 'input_bid']
        if is_db:
            db_ini_path = self.create_db_ini(upload_json_path)
        else:
            db_ini_path = None
        cmd = self.trans.create_cmd(cmd_params, db_ini_path, engine_type, server_ip, server_port)

        return run_cmd(cmd, flag=True, logger=self.logger)

    def load_db_config(self, db_config_path=None):
        if not bool(db_config_path) or not os.path.exists(db_config_path):
            db_config_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "db_config.ini"))
        conf = read_ini_config(db_config_path)
        on = conf.get("DATABASE_CONFIG", "on")
        type = conf.get("DATABASE_CONFIG", "type")
        db_path = conf.get("DATABASE_CONFIG", "db_path")
        host = conf.get("REDIS", "host")
        port = conf.get("REDIS", "port")
        password = conf.get("REDIS", "password")
        table_index = conf.get("REDIS", "table_index")
        timeout = conf.get("REDIS", "timeout")
        temporary = conf.get("SQLITE", "temporary")
        transfer_log_path = conf.get("TRANSFER_LOG_PATH", "transfer_log_path")

        database_config = {
            "on": on,
            "type": type,
            "db_path": db_path
        }
        redis_config = {
            "host": host,
            "port": int(port),
            "password": password,
            "table_index": table_index,
            "timeout": timeout
        }
        sqlite_config = {
            "temporary": temporary
        }

        return transfer_log_path, redis_config, sqlite_config, database_config

    def multi_thread_upload(self, upload_pool, thread_num=10):
        """muti thread upload resource.

        Args:
            upload_pool (list or tuple): Store a list or ancestor of uploaded files.
            thread_num (int): Number of threads, 10 threads are enabled by default.

        """
        threads = []
        for _ in range(thread_num):
            t = Thread(target=self.upload_asset, args=(upload_pool.pop(),))
            threads.append(t)
            t.start()
        for t in threads:
            t.join()

    def thread_pool_upload(self, upload_pool, pool_size=10):
        """Thread pool upload.

        Args:
            upload_pool (list or tuple): store a list or ancestor of uploaded files.
            pool_size (int): thread pool size, default is 10 threads.

        """
        pool = ThreadPoolExecutor(pool_size)
        for i in range(len(upload_pool)):
            pool.submit(self.upload_asset, upload_pool[i])
        pool.shutdown(wait=True)



# -*- coding: utf-8 -*-
"""Upload demo."""
from rayvision_api import RayvisionAPI
from rayvision_api.utils import append_to_upload
from rayvision_sync.upload import RayvisionUpload
from rayvision_sync.utils import cutting_upload

api = RayvisionAPI(access_id="xxxx",
                   access_key="xxxx",
                   domain="task.renderbus.com",
                   platform="21")

config_list = [
    r"C:\workspace\work\tips.json",
    r"C:\workspace\work\task.json",
    r"C:\workspace\work\asset.json",
    r"C:\workspace\work\upload.json",
]

CONFIG_PATH = {
    "tips_json_path": r"C:\workspace\work\tips.json",
    "task_json_path": r"C:\workspace\work\task.json",
    "asset_json_path": r"C:\workspace\work\asset.json",
    "upload_json_path": r"C:\workspace\work\upload.json",
}

UPLOAD = RayvisionUpload(api)
custom_info_to_upload = [
    r"D:\houdini\CG file\local"
]

# custom_info_to_upload = r"D:\houdini\CG file\katana_file"
# append_to_upload(custom_info_to_upload, r"D:\test\upload.json")
# UPLOAD.upload_asset(upload_json_path=r"D:\myproject\local_internal\upload.txt",
#                     transmit_type="upload_json",
#                     engine_type='aspera',
#                     server_ip="45.251.92.16",
#                     server_port="12121")

# UPLOAD.upload_config(task_id="5165465",
#                      config_file_list=config_list,
#                      server_ip="45.251.92.16",
#                      server_port="12121")


# cut upload according to the number of resources
# upload_pool = cutting_upload(r"D:\test\test_upload\1586250829\upload.json", max_resources_number=800)
# print(upload_pool)


# Multi-thread upload
# UPLOAD.multi_thread_upload(upload_pool, thread_num=20)

# Thread pool upload
# UPLOAD.thread_pool_upload(upload_pool, pool_size=20)
#

upload_method = 2
if upload_method == 1:
#     # step4.1:Json files are uploaded in conjunction with CG resources
    UPLOAD.upload(task_id="41235091",
                  engine_type='aspera',
                  server_ip="45.251.92.16",
                  server_port="12121",
                  task_json_path=r"C:\workspace\work\task.json",
                  tips_json_path=r"C:\workspace\work\tips.json",
                  asset_json_path=r"C:\workspace\work\asset.json",
                  upload_json_path=r"C:\workspace\work\upload.json")
elif upload_method == 2:
    # step4.2:CG resource files and json are uploaded separately
    UPLOAD.upload_asset(upload_json_path=CONFIG_PATH["upload_json_path"], engine_type='')
    # UPLOAD.upload_config(str(41235091), list(CONFIG_PATH.values()))



"""CG errors."""

class RayvisionError(Exception):
    """Raise RayvisionError if something wrong."""

    def __init__(self, error_code, error, *args, **kwargs):
        """Initialize error message, inherited Exception.

        Args:
            error_code (int): Error status code.
            error (str): Error message.
            args (set): Other parameters.
            kwargs (dict): Other keyword parameters.

        """
        super(RayvisionError, self).__init__(self, error, *args, **kwargs)
        self.error_code = error_code
        self.error = error


class UnsupportedDatabaseError(Exception):
    """The local database supports only Redis and Sqlite"""

class UnsupportedEngineType(Exception):
    """Engine Only Support aspera and raysync"""


"""The plugin of the pytest.

The pytest plugin hooks do not need to be imported into any test code, it will
load automatically when running pytest.

References:
    https://docs.pytest.org/en/2.7.3/plugins.html

"""

import os
import sys

# pylint: disable=import-error
import pytest

from rayvision_api.connect import Connect
from rayvision_api.core import RayvisionAPI
from rayvision_sync.download import RayvisionDownload
from rayvision_sync.manage import RayvisionManageTask
from rayvision_sync.transfer import RayvisionTransfer
from rayvision_sync.upload import RayvisionUpload


@pytest.fixture()
def user_trans_info():
    """Get user trans info."""
    return {
        'config_bid': '54252',
        'input_bid': '46456',
        'output_bid': '20214',
        'user_id': '100093088',
        'domain': 'tasks.renderbus.com',
        'platform': '2',
        'local_os': 'windows',
    }


@pytest.fixture(name='user_info_dict')
def user_info():
    """Get the user's login information."""
    return {
        'domain': 'task.renderbus.com',
        'platform': '2',
        'access_id': 'test_access_id',
        'access_key': 'test_access_key',
        'protocol': 'https'
    }


@pytest.fixture()
def expected_result():
    """Get the expected_result information."""
    return [
        {
            'task_id': '6419169',
            'sub_task_status': [],
            'task_status_description': 'Done(with failed frame)',
            'is_opener': '0',
            'task_status_text': 'render_task_status_0',
            'task_status_code': '30',
            'output_file_name': 'test_info_1'
        },
        {
            'task_id': '647611',
            'sub_task_status': [
                {
                    'task_id': '6416355',
                    'sub_task_status': [],
                    'task_status_description': 'Failed',
                    'is_opener': '1',
                    'task_status_text': 'render_task_status_-2',
                    'task_status_code': '45',
                    'output_file_name': 'test_info_3'
                }
            ],
            'task_status_description': 'Abort',
            'is_opener': '1',
            'task_status_text': 'render_task_status_-1',
            'task_status_code': '35',
            'output_file_name': 'test_info_2'
        }
    ]


@pytest.fixture()
def api(user_info_dict, mocker, tmpdir):
    """Create connect API object."""
    mocker_task_id = mocker.patch.object(Connect, "post")
    mocker_task_id.return_value = {}
    rayvision_api = RayvisionAPI(**user_info_dict)
    if "win" in sys.platform.lower():
        local_os = "windows"
        os.environ["USERPROFILE"] = str(tmpdir)
    else:
        local_os = "linux"
        os.environ["HOME"] = str(tmpdir)
    rayvision_api.user._info = {
        'config_bid': "10101",
        'input_bid': "20202",
        "output_bid": "20202",
        "domain": "task.renderbus.com",
        "platform": "2",
        "local_os": local_os,
        "user_id": "1015646",
    }
    return rayvision_api


@pytest.fixture()
def manage(api):
    """Create an RayvisionManageTask object."""
    return RayvisionManageTask(api.query)


@pytest.fixture()
def rayvision_transfer(user_trans_info, manage):
    """Create an RayvisionTransfer object."""
    return RayvisionTransfer(manage_task=manage, **user_trans_info)


@pytest.fixture()
def rayvision_download(api):
    """Create an RayvisionDownload object."""
    return RayvisionDownload(api)


@pytest.fixture()
def rayvision_upload(api):
    """Create an RayvisionDownload object."""
    return RayvisionUpload(api)



"""Test the rayvison_sync.rayvision_transfer functions."""

# pylint: disable=import-error
import pytest


def test_transport_info(rayvision_transfer):
    """Test we can get transport info."""
    assert sorted(list(rayvision_transfer.transport_info.keys())) == [
        'aspera', 'raysync']


def test_create_cmd(rayvision_transfer):
    """Test we can get a str."""
    cmd_params = [
        "upload_path",
        "D:/gitlab_demo/temp/demo_bak/rayvision/help/scenes/demo_scenc.mb",
        "/D/gitlab_demo/temp/demo_bak/rayvision/help/scenes/demo_scenc.mb",
        "1048576",
        "false", "config_bid"
    ]
    data = rayvision_transfer.create_cmd(cmd_params)
    # noqa: E501  # pylint: disable=line-too-long
    result = '-E "aspera" -H "45.251.92.29" -P "10221" -S "54252" -U "100093088" -T "upload_path" ' \
             '-L "D:/gitlab_demo/temp/demo_bak/rayvision/help/scenes/demo_scenc.mb" ' \
             '-R "/D/gitlab_demo/temp/demo_bak/rayvision/help/scenes/demo_scenc.mb" ' \
             '-r "2" -K "false" -s "1048576" -C "None"'
    assert result in data


@pytest.mark.parametrize("transports_json, domain_name, platform", [
    (None, None, '2'),
    (None, "jop.foxrenderfarm.com", '2'),
    (None, "task.renderbus.com", '2'),
    (None, "jop.test.com", '2'),
])
def test_parse_transports_json(rayvision_transfer, transports_json,
                               domain_name, platform):
    """Test we can get correct data."""
    result = rayvision_transfer.parse_transports_json(transports_json,
                                                      domain_name, platform)
    assert sorted(list(result.keys())) == ['aspera', 'raysync']


# -*- coding: utf-8 -*-
"""Download demo."""

# Import local modules
from rayvision_api import RayvisionAPI
from rayvision_sync.download import RayvisionDownload

api = RayvisionAPI(access_id="xxxx",
                   access_key="xxx",
                   domain="task.renderbus.com",
                   platform="6")

download = RayvisionDownload(api)
# download.download(download_filename_format="true", server_path="40112885_muti_layer_test")
# download.auto_download([41372307], download_filename_format="false")
# download.auto_download_after_task_completed([41372307], download_filename_format="true", engine_type="raysync")



"""Test the rayvision_sync download functions."""

# pylint: disable=import-error
import pytest

from rayvision_sync.manage import RayvisionManageTask


@pytest.mark.parametrize('task_id_list,max_speed,sleep_time', [
    ([258951, 465400, 4664165], '1024', 0),
    ([111], '10240', 2),
    ([8951, 6144, 4665], '2048', 100)
])
def test_download(rayvision_download, task_id_list, max_speed, sleep_time,
                  expected_result, mocker):
    """Test output_file_names, we can get a expected result."""
    mocker_task_id = mocker.patch.object(RayvisionManageTask,
                                         'get_task_status')
    mocker_task_id.return_value = expected_result
    result = rayvision_download.download(task_id_list, max_speed, sleep_time)
    assert result is True


@pytest.mark.parametrize('task_id_list,max_speed,sleep_time', [
    ([25, 614654, 4664165], '1024', 0),
    ([11], '10240', 12),
    ([8951, 6144, 4665], '2048', 100)
])
def test_auto_download(rayvision_download, task_id_list, max_speed, sleep_time,
                       expected_result, mocker):
    """Test auto_download, we can get a expected result."""
    mocker_task_id = mocker.patch.object(RayvisionManageTask,
                                         'get_task_status')
    mocker_task_id.return_value = expected_result
    result = rayvision_download.auto_download(task_id_list, max_speed,
                                              sleep_time)
    assert result is True


@pytest.mark.parametrize('task_id_list,max_speed,sleep_time', [
    ([11], '10240', 20),
    ([8951, 6144, 4665], '2048', 5)
])
def test_auto_download_after_task_completed(rayvision_download, task_id_list,
                                            max_speed, sleep_time,
                                            expected_result, mocker):
    """Test auto_download_after_task_completed, we can get a expected result."""
    mocker_task_id = mocker.patch.object(RayvisionManageTask,
                                         'get_task_status')
    mocker_task_id.return_value = expected_result
    result = rayvision_download.auto_download_after_task_completed(
        task_id_list, max_speed, sleep_time)
    assert result is True



# -*- coding: utf-8 -*-
"""Test the rayvision_sync utils functions."""

from builtins import str

# pylint: disable=import-error
import pytest

from rayvision_sync import utils


@pytest.mark.parametrize("handle_str, str_decode", [
    ("mxxx", "utf-8"),
    ("天天", "utf-8"),
    (b"mxxx", "gbk"),
])
def test_str2unicode(handle_str, str_decode):
    """Test we can get a string."""
    result = utils.str2unicode(handle_str, str_decode)
    assert isinstance(result, str)


@pytest.mark.parametrize("flag, returncode, err_messages", [
    (bool(1), 1, ['error', 'debug']),
])
def test_handle_cmd_result(flag, returncode, err_messages):
    """Test process cmd results."""
    result = utils.handle_cmd_result(flag, returncode, err_messages)
    assert isinstance(result, int)


def test_run_cmd():
    """Test run cmd."""
    cmd_str = "dir"
    result = utils.run_cmd(cmd_str)
    assert isinstance(result, int)


@pytest.mark.parametrize("task_status_code,language ", [
    ("0", "1"),
    ("5", "0"),
    ("35", "1"),
])
def test_get_task_status_description(task_status_code, language):
    """Test task status description."""
    result = utils.get_task_status_description(task_status_code, language)
    if task_status_code == "0":
        assert result == "Waiting"
    elif task_status_code == "5":
        assert result == "渲染中"
    else:
        assert result == "Abort"



# -*- coding: utf-8 -*-
"""Transfer Module.

Execute the cmd command to make the last asset and
download assets.

"""

# Import built-in modules
from __future__ import print_function

import codecs
import json
import logging
import os
import sys

from rayvision_sync.constants import ENGINE_TYPE
from rayvision_sync.exception import UnsupportedEngineType


class RayvisionTransfer(object):
    """Transfer including upload files and download files."""

    def __init__(self, config_bid, input_bid, domain, platform, local_os,
                 user_id, output_bid=None, manage_task=None):
        """Initialize the configuration of the transfer.

        Args:
            config_bid (str): transport configuration id.
            input_bid (str): storage id.
            output_bid (str): storage id.
            domain (str): domain name, like task.renderbus.com".
            platform (str): platform id, for example: "2".
            local_os (str): system name,  Only support "window" or "linux".
            user_id (str): user accound id.
            manage_task (RayvisionManageTask, optional): Instantiated object
                of the management tasks, If it is just uploading, this
                parameter can not be passed. If it is downloaded, this
                parameter must have.
        """
        self.logger = logging.getLogger(__name__)
        self.config_bid = config_bid
        self.input_bid = input_bid
        self.output_bid = output_bid
        self.domain = domain
        self.platform = platform
        self.local_os = local_os
        self.user_id = user_id
        self.manage_task = manage_task
        self.user_info = {
            'config_bid': self.config_bid,
            'input_bid': self.input_bid,
            'output_bid': self.output_bid,
            'user_id': self.user_id,
            'domain': self.domain,
            'platform': self.platform,
            'local_os': self.local_os,
        }
        current_dir = os.path.dirname(os.path.realpath(__file__))
        self.transmitter_exe = self.init_transmitter(current_dir, self.local_os)

        self._transports_json = os.path.join(current_dir, 'transmission',
                                             'transports.json')
        self.transport_info = self.parse_transports_json()

    @staticmethod
    def init_transmitter(current_dir, local_os):
        """Gets the path of the transfer software.

        Args:
            current_dir: transfer base directory.
            local_os: system name, Only support "window" or "linux"

        Returns: transfer software absolute path.

        """
        if local_os == 'windows':
            transmitter_exe = os.path.join(current_dir, 'transmission',
                                           local_os,
                                           'rayvision_transmitter.exe')
        else:
            transmitter_exe = os.path.join(current_dir, 'transmission',
                                           "linux",
                                           'rayvision_transmitter')
        return transmitter_exe

    def parse_transports_json(self, transports_json=None, domain=None,
                              platform=None):
        """Analyze transports.json,obtain transfer server info.

        Extract the transmission configuration information of the
        corresponding platform in transports.json.

        Args:
            transports_json (str, optional): Path to transfer configuration
                files.
            domain (str, optional): Domain name.
            platform (str, optional): Platform number.

        Returns:
            dict: Transfer configuration information
                .e.g:
                    {
                        "engine_type":"aspera",
                        "server_name":"HKCT",
                        "server_ip":"render-client.raysync.cn",
                        "server_port":"10221"
                    }

        """
        if not domain:
            domain = self.domain
        if not transports_json:
            transports_json = self._transports_json
        if 'foxrenderfarm' in domain:
            key_first_half = 'foxrenderfarm'
        else:
            key_first_half = 'renderbus'

        if not platform:
            platform = self.platform
        key_second_half = self._get_key_second_half(platform)

        if key_second_half == 'default':
            key = key_second_half
        else:
            key = '%s_%s' % (key_first_half, key_second_half)

        with codecs.open(transports_json, 'r', 'utf-8') as f_transports:
            transports_info = json.load(f_transports)
        return transports_info[key]

    @staticmethod
    def _get_key_second_half(platform):
        """Get the key corresponding to the platform number.

        Returns:
             key_second_half(str): Representative submits tasks to several
                platforms.

        """
        setting_mappings = {
            '2': 'www2',
            '3': 'www3',
            '6': 'www6',
            '20': 'pic',
            '21': 'gpu',
            '35': 'gpu5',
        }
        return setting_mappings[platform]

    def create_cmd(self, cmd_params, db_ini_path=None, engine_type="aspera", server_ip=None, server_port=None):
        """Splice a cmd command.

        Args:
            cmd_params (list): Parameters required by the cmd command.
                Examples::

                    [
                        transmit_type, # Transmission type
                        local_path, # Local file path
                        output_file_name, # File path uploaded to the server
                        max_speed, # Maximum transfer speed of upload
                        'true', # Whether to save the path
                        'output_bid', # Transfer id
                    ]

            db_ini_path (str): Database path.
            engine_type (str, optional): set engine type, support "aspera" and "raysync", Default "aspera".
            server_ip (str, optional): transmit server host,
                if not set, it is obtained from the default transport profile.
            server_port (str, optional): transmit server port,
                if not set, it is obtained from the default transport profile.

        Returns:
            str: Cmd command.

        """
        if not sys.platform.startswith('win'):
            os.environ["LD_LIBRARY_PATH"] = os.path.dirname(self.transmitter_exe)
            chmod_str = "chmod 777 -R {}/*".format(os.path.dirname(self.transmitter_exe))
            os.system(chmod_str)
        if not bool(engine_type):
            engine_type = "aspera"
        if engine_type not in ENGINE_TYPE:
            msg = "{} is not a supported transport engine, " \
                  "currently only support 'aspera' and 'raysync'".format(engine_type)
            raise UnsupportedEngineType(msg)
        transmit_cmd = ('echo y|"{exePath}" -E "{engineType}"'
                        ' -H "{serverIp}" -P "{serverPort}" -S "{download_id}"'
                        ' -U "{userId}" -T "{transmit_type}" -L "{local_path}"'
                        ' -R "{server_path}" -r "{maxConnectFailureCount}"'
                        ' -K "{keep_path}" -s "{max_speed}" -C "{database_config_path}"'
                        ' ').format(
            exePath=self.transmitter_exe,
            engineType=engine_type,
            serverIp=server_ip if server_ip else self.transport_info[engine_type]['server_ip'],
            serverPort=server_port if server_port else self.transport_info[engine_type]['server_port'],
            download_id=self.user_info[cmd_params[5]],
            userId=self.user_id,
            transmit_type=cmd_params[0],
            local_path=cmd_params[1],
            server_path=cmd_params[2],
            maxConnectFailureCount='2',  # default is 2.
            keep_path=cmd_params[4],
            max_speed=cmd_params[3],
            database_config_path=db_ini_path)
        return transmit_cmd



"""Test the rayvision_sync manage functions."""

# pylint: disable=import-error
import pytest

from rayvision_api.operators import QueryOperator


@pytest.fixture()
def task_status_list():
    """Get the task_status_list information."""
    return [
        {
            "task_status_code": "-1",
            "is_opener": "1",
            "output_file_name": "test1",
            "sub_task_status": [
                {
                    "task_status_code": "11",
                    "output_file_name": "test2",
                    "is_opener": "0",
                }
            ]
        },
        {
            "task_status_code": "10",
            "is_opener": "0",
            "output_file_name": "test3",
            "sub_task_status": []
        },
    ]


@pytest.fixture()
def task_info_list():
    """Get the task_info_list information."""
    return [
        {
            "id": 6419169,
            "taskStatus": 30,
            "statusText": "render_task_status_0",
            "isOpen": 0,
            "outputFileName": "test_info_1",
        },
        {
            "id": 647611,
            "taskStatus": 35,
            "statusText": "render_task_status_-1",
            "isOpen": 1,
            "outputFileName": "test_info_2",
            "respRenderingTaskList": [
                {
                    "id": 6416355,
                    "taskStatus": 45,
                    "statusText": "render_task_status_-2",
                    "isOpen": 1,
                    "outputFileName": "test_info_3",
                }
            ],
        }
    ]


def test_find_task_status_codes(manage, task_status_list):
    """Test find_task_status_codes, we can get a expected result."""
    result = manage.find_task_status_codes(task_status_list)
    assert result == ['11', '10']


def test_output_file_names(manage, task_status_list):
    """Test output_file_names, we can get a expected result."""
    result = manage.output_file_names(task_status_list)
    assert result == ['test2', 'test3']


def test_task_info_iterater(manage, task_info_list, expected_result):
    """Test task_info_iterater, we can get a expected result."""
    data = manage.task_info_iterater(task_info_list)
    assert data == expected_result


@pytest.mark.parametrize('task_id', [
    159423,
    597316,
    2164164
])
def test_is_task_end(manage, mocker, task_info_list, task_id):
    """Test is_task_end, we can get a expected result."""
    task_info = {
        'items': task_info_list
    }
    mocker_task_id = mocker.patch.object(QueryOperator, 'task_info')
    mocker_task_id.return_value = task_info
    result = manage.is_task_end(task_id)
    assert isinstance(result, bool)



"""Describe our module distribution to Distutils."""
from setuptools import find_packages, setup


def parse_requirements(filename):
    with open(filename, "r") as f:
        for line in f:
            yield line.strip()


setup(
    name="rayvision_sync",
    author="Shenzhen Rayvision Technology Co., Ltd",
    author_email="developer@rayvision.com",
    url="https://github.com/renderbus/rayvision_sync",
    package_dir={"": "."},
    packages=find_packages("."),
    description=("Upload configuration file and asset file, download result "
                 "file."),
    entry_points={},
    install_requires=list(parse_requirements("requirements.txt")),
    package_data={
        'rayvision_sync': ["./transmission/*", "./transmission/*/*", "./transmission/*/*/*", "./transmission/*/*/*/*", "*.ini"],
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
)



"""Transfer the required public functions."""

# Import built-in modules
from __future__ import print_function

import codecs
import json
import logging
import os
import subprocess
import sys
import time

import configparser
from builtins import str

# Import local models
from rayvision_sync.constants import TASK_STATUS_DESCRIPTION
# Import third-party modules
from rayvision_sync.exception import RayvisionError


def print_to_log(cmd, logger):
    """Output the CMD command information.

    If there is an error in the log, it will be recorded in a list.

    Args:
        cmd (subprocess.Popen): The object obtained by executing the cmd
            command.
        logger (func): The log function.
            .e.g:
                SDK_LOG.info.

    Returns:
        list: All errors output by the cmd command.

    """
    err_messages = []
    while True:
        result_line = cmd.stdout.readline()
        if result_line:
            result_line = result_line.strip()
            if result_line:
                if logger:
                    result_line = str2unicode(result_line)
                    logger(result_line)
                    if ('upload file fail' in result_line or
                            ('get path' in result_line and
                             'info failed' in result_line)):
                        err_messages.append(result_line)
        else:
            break
    return err_messages


def handle_cmd_result(flag, returncode, err_messages, logger=None):
    """Process cmd results.

    Perform the corresponding operation according to the status code.
    0 means success, 11 means unable to retry, 10 means retry,9 means
    parameter or domain name resolution error.

    Args:
        flag (bool): Used to identify uploads.
        returncode (int): The status code returned by cmd after execution.
        err_messages (list of str): List of error messages.
        logger (logging.info or logging.debug): Log output object method.

    Returns:
        int: Status code.

    """
    logger = logger or logging.getLogger(__name__)
    if flag and returncode == 0:
        pass
    elif flag and returncode == 11:
        for err_message in err_messages:
            logger("err_message: %s", err_message)
        logger("cmdp.returncode: %s", returncode)
        logger("Upload failed, there has non-uploadable files")

    elif flag and returncode == 10:
        for err_message in err_messages:
            logger("err_message: %s", err_message)
        logger("cmdp.returncode: %s", returncode)
        logger("Err_message has files that have not been uploaded"
               " successfully, will enter retry upload ")

    elif flag and returncode == 9:
        for err_message in err_messages:
            logger("err_message: %s", err_message)
        logger("cmdp.returncode: %s", returncode)
        logger("Parameter or domain name resolution error.")

    return returncode


def run_cmd(cmd_str, my_shell=True, print_log=True, flag=None, logger=None):
    """Run cmd.

    If the cmd runs with an error, it will print the error message and return
    to False.

    Args:
        cmd_str (str): String of the cmd command.
        my_shell (bool, optional): Accept a string type variable as a command
            and call the shell to execute the string.
        print_log (bool, optional): Print log, True: print, False: not print.
        flag (bool, optional): default is None.
        logger (logging, optional): Log object.

    Returns:
        bool: True is success, False is wrong.

    """
    if not logger:
        logger = logging.getLogger(__name__)

    if print_log:
        logger = logger.info
    else:
        logger = logger.debug

    logger(u'cmd...%s', str2unicode(cmd_str))

    if sys.version_info[0] == 2:
        cmd_str = str2unicode(cmd_str)
        cmd_str = cmd_str.encode(sys.getfilesystemencoding())


    cmd_result = subprocess.Popen(cmd_str, stdin=subprocess.PIPE,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  shell=my_shell)

    err_messages = print_to_log(cmd_result, logger)
    cmd_result.communicate()
    returncode = cmd_result.returncode
    return handle_cmd_result(flag, returncode, err_messages, logger)


def upload_retry(func):
    """Determine whether to retry according to the status code of cmd."""

    def inner(*args, **kwargs):
        """Handle."""
        time_interval = 5
        times = 0
        logger = logging.getLogger(__name__)
        while True:
            result = func(*args, **kwargs)
            if result == 0:
                return True
            if times > 10:
                if result == 11:
                    raise RayvisionError(200025, "There are files that cannot"
                                                 "be uploaded, please check the file is exists.")
                if result == 10:
                    time.sleep(time_interval)
                    time_interval += 10
                    if time_interval > 600:
                        time_interval = 600
                    logger.info("Retrying upload......")
                elif result == 9:
                    raise Exception("Parameter or domain name resolution"
                                    "error.")
                elif result in [1, 2, 3, 4, 5, 6, 7, 8]:
                    raise RayvisionError(200024, "argument format invalid"
                                                 "please check the argument.")
            else:
                times += 1
                logger.info("Retrying upload......")

    return inner


def str2unicode(handle_str, str_decode='default'):
    """Decode the string.

    Args:
        handle_str (str): String to be decoded.
        str_decode (str, optional): What encoding needs to be decoded,
            the default is lowercase.

    Returns:
        str: Need to decode the string.

    """
    if not isinstance(handle_str, str):
        try:
            if str_decode != 'default':
                handle_str = handle_str.decode(str_decode.lower())
            else:
                try:
                    handle_str = handle_str.decode('utf-8')
                except (AttributeError, UnicodeError):
                    try:
                        handle_str = handle_str.decode('gbk')
                    except (AttributeError, UnicodeError):
                        handle_str = handle_str.decode(
                            sys.getfilesystemencoding())
        except (AttributeError, UnicodeError) as err:
            logger = logging.getLogger(__name__)
            logger.warning('str2unicode:decode failed')
            logger.warning(str(err))
    return handle_str


def get_task_status_description(task_status_code=None, language='1'):
    """Get task status description.

    Get the status of the task now by the task status code.

    Args:
        task_status_code (str): Task status code.
        language (str): Language, "0": Simplified Chinese; "1": English

    Returns:
        str: Task status information.
            e.g.:
                "Waiting".

    Raises:
        RayvisionError: Task status code is None.

    """
    try:
        task_status_description = TASK_STATUS_DESCRIPTION.get(
            task_status_code, {})[language]
        return task_status_description
    except KeyError:
        logger = logging.getLogger(__name__)
        logger.error('Get empty task_status_description,'
                     ' Please check the input.')
        raise RayvisionError(1000000, "Get empty task_status_description,"
                                      " Please check the input")


def create_transfer_params(api):
    """Takes a parameter from the authentication object.

    Args:
        api: Objects that have been authenticated.

    Returns:
        dict

    """
    params = {
        'config_bid': api.user_info['config_bid'],
        'input_bid': api.user_info['input_bid'],
        "output_bid": api.user_info["output_bid"],
        "domain": api.user_info['domain'],
        "platform": api.user_info['platform'],
        "local_os": api.user_info['local_os'],
        "user_id": api.user_info['user_id'],
    }
    return params


def read_ini_config(db_config_path):
    """read ini file"""
    conf = configparser.ConfigParser()
    conf.read(db_config_path, encoding="utf-8")
    return conf


def json_load(json_path, encoding='utf-8'):
    """Load the data from the json file.

    Args:
        json_path (str): Json file path.
        encoding (str): Encoding, default is ``utf-8``.

    Returns:
        dict: Data in the json file.
            e.g.:
                {
                    "task_info"
                }

    """
    if os.path.exists(json_path):
        with codecs.open(json_path, 'r', encoding=encoding) as f_json:
            data = json.load(f_json)

        return data


def write_json(path, info):
    """write json file."""
    with codecs.open(path, 'w', 'utf-8') as f_tips_json:
        json.dump(info, f_tips_json, indent=4, ensure_ascii=False)


def cutting_upload(upload_path, max_resources_number=None, after_cutting_position=None):
    """Cut upload.json according to the number of custom files.

    Args:
        upload_path (str): upload.json absolute path.
        max_resources_number (int): Maximum number of resources in each upload file.
        after_cutting_position (str): save location of upload file generated after cutting.

    Returns:
        list: Absolute path of all upload files generated after cutting, excluding original upload files.
            e.g.:
                ['D:\\test\\test_upload\\1586250829\\upload_1.json',
                'D:\\test\\test_upload\\1586250829\\upload_2.json']

    """
    if not os.path.exists(upload_path):
        raise RayvisionError(200001, "upload file is not exist: {}".format(upload_path))
    if max_resources_number is None:
        return upload_path
    if after_cutting_position is None or not os.path.exists(after_cutting_position):
        after_cutting_position = os.path.dirname(upload_path)

    upload_info = json_load(upload_path)
    asset = upload_info.get("asset", [])
    if max_resources_number >= len(asset):
        return
    count = 1
    cut_json_pool = []
    for per_index in range(0, len(asset) + 1, max_resources_number):
        if per_index == 0:
            cut_per_info = asset[0:max_resources_number]
        else:
            cut_per_info = asset[per_index:per_index + max_resources_number]
        cut_info = {"asset": cut_per_info}
        to_path = os.path.normpath(os.path.join(after_cutting_position, "upload_{}.json".format(str(count))))
        write_json(to_path, cut_info)
        count += 1
        cut_json_pool.append(to_path)
    return cut_json_pool



"""Download models.

Including download, automatic download (the task is fully rendered,
one frame is downloaded after downloading one frame), download so
the un-downloaded task is recorded (the task is rendered).

"""

import os
# Import built-in modules
import time

from rayvision_sync.manage import RayvisionManageTask
# Import local modules
from rayvision_sync.transfer import RayvisionTransfer
from rayvision_sync.utils import create_transfer_params
from rayvision_sync.utils import run_cmd
from rayvision_sync.utils import str2unicode


class RayvisionDownload(object):
    """Downloader.

    Download all the passed tasks by calling the cmd command.

    """

    def __init__(self, api):
        """Initialize instance."""
        params = create_transfer_params(api)
        self.api = api
        self.trans = RayvisionTransfer(**params)
        self.manage_task = self.trans.manage_task or RayvisionManageTask(api.query)
        self.logger = self.trans.logger

    def _download_log(self, task_id_list, local_path):
        """Download log Settings.

        Args:
            task_id_list (list): List of tasks ids that need to be downloaded.

        """
        self.logger.info('INPUT:')
        self.logger.info('=' * 20)
        self.logger.info('task_id_list: %s', task_id_list)
        self.logger.info('local_path: %s', local_path)
        self.logger.info('=' * 20)

    def _check_local_path(self, local_path):
        """Check the download path.

        Args:
            local_path (str):  Local path to download.

        Returns:
            str

        """
        if not local_path:
            if self.api.user_info['local_os'] == "windows":
                local_path = os.path.join(os.environ["USERPROFILE"],
                                          "renderfarm_sdk")
            else:
                local_path = os.path.join(os.environ["HOME"], "renderfarm_sdk")
        return local_path

    @staticmethod
    def check_params(task_id_list, custom_server_output_path):
        """Check the parameters.
        
        Task_id_list and custom_server_output_path must have one.
        """
        if not task_id_list and not custom_server_output_path:
            raise Exception("One of the task_id_list and custom_server_output_path"
                            " must exist")

    def download(self, task_id_list=None,
                 max_speed=None, print_log=True,
                 download_filename_format="true",
                 local_path=None, server_path=None,
                 engine_type="aspera", server_ip=None, server_port=None):
        """Download and update the undownloaded record.

        Args:
            task_id_list (list of int): List of tasks ids that need to be
                downloaded.
            max_speed (str, optional): Download speed limit,
                The unit of ``max_speed`` is KB/S,default value is 1048576
                KB/S, means 1 GB/S.
            print_log (bool, optional): Print log, True: print, False: not
                print.
            download_filename_format: File download local save style,
                "true": tape task ID and scene name,
                "false" : download directly without doing processing.
            local_path (str): Download file locally save path,
                default Window system is "USERPROFILE" environment variable address splicing "renderfarm_sdk",
                Linux system is "HOME" environment variable address splicing "renderfarm_sdk",
            server_path (str or list): The user customizes the file structure to be downloaded from
                the output server, and all file structures are downloaded by default,
                example: "18164087_test/l_layer".
            engine_type (str, optional): set engine type, support "aspera" and "raysync", Default "aspera".
            server_ip (str, optional): transmit server host,
                if not set, it is obtained from the default transport profile.
            server_port (str, optional): transmit server port,
                if not set, it is obtained from the default transport profile.

        Returns:
            bool: True is success.

        """
        self.check_params(task_id_list, server_path)
        local_path = self._check_local_path(local_path)
        self.logger.info("[Rayvision_sync start download .....]")
        self._download_log(task_id_list, local_path)

        self._run_download(task_id_list, local_path, max_speed, print_log,
                           download_filename_format, server_path,
                           engine_type=engine_type, server_ip=server_ip, server_port=server_port)
        self.logger.info("[Rayvision_sync end download.....]")
        return True

    def auto_download(self, task_id_list=None, max_speed=None,
                      print_log=False, sleep_time=10,
                      download_filename_format="true",
                      local_path=None,
                      engine_type="aspera", server_ip=None, server_port=None):
        """Automatic download (complete one frame download).

        Wait for all downloads to update undownloaded records.

        Args:
            task_id_list (list of int): List of tasks ids that need to be
                downloaded.
            max_speed (str, optional): Download speed limit,
                The unit of 'max_speed' is KB/S,default value is 1048576 KB/S,
                means 1 GB/S.
            print_log (bool, optional): Print log, True: print, False: not
                print.
            sleep_time (int, optional): Sleep time between download,
                unit is second.
            download_filename_format: File download local save style,
                "true": tape task ID and scene name,
                "false" : download directly without doing processing.
            local_path (str): Download file locally save path,
                default Window system is "USERPROFILE" environment variable address splicing "renderfarm_sdk",
                Linux system is "HOME" environment variable address splicing "renderfarm_sdk".
            engine_type (str, optional): set engine type, support "aspera" and "raysync", Default "aspera".
            server_ip (str, optional): transmit server host,
                if not set, it is obtained from the default transport profile.
            server_port (str, optional): transmit server port,
                if not set, it is obtained from the default transport profile.

        Returns:
            bool: True is success.

        """
        local_path = self._check_local_path(local_path)
        self.logger.info("[Rayvision_sync start auto_download.....]")
        self._download_log(task_id_list, local_path)

        self._auto_download_tool(task_id_list, sleep_time,
                                 max_speed, print_log, local_path,
                                 download_filename_format,
                                 engine_type=engine_type, server_ip=server_ip, server_port=server_port)
        self.logger.info("[Rayvision_sync end auto_download.....]")
        return True

    def _auto_download_tool(self, task_id_list, sleep_time,
                            max_speed, print_log, local_path,
                            download_filename_format="true",
                            engine_type=None, server_ip=None, server_port=None):
        """Automatic download (complete one frame download).

        Args:
            task_id_list(list of int): List of tasks ids that need to be
                downloaded.
            sleep_time(int):  Sleep time between download.
            max_speed(str): Download speed limit.
            print_log(bool): Print log, True: print, False: not print.
            download_filename_format(str): File download local save style,
                "true": tape task ID and scene name,
                "false" : download directly without doing processing.

        """
        while True:
            if task_id_list:
                time.sleep(float(sleep_time))
                for task_id in task_id_list:
                    is_task_end = self.manage_task.is_task_end(task_id)
                    self._run_download([task_id], local_path, max_speed,
                                       print_log, download_filename_format,
                                       engine_type=engine_type, server_ip=server_ip, server_port=server_port)

                    if is_task_end is True:
                        self.logger.info('The tasks end: %s', task_id)
                        task_id_list.remove(task_id)
            else:
                break

    def auto_download_after_task_completed(self, task_id_list=None,
                                           max_speed=None, print_log=True,
                                           sleep_time=10,
                                           download_filename_format="true",
                                           local_path=None,
                                           engine_type="aspera", server_ip=None, server_port=None):
        """Auto download after the tasks render completed.

        Args:
            task_id_list(list of int): List of tasks ids that need to be
                downloaded.
            max_speed(str, optional): Download speed limit,
                The unit of 'max_speed' is KB/S,default value is 1048576 KB/S,
                means 1 GB/S.
            print_log(bool, optional): Print log, True: print, False: not
                print.
            sleep_time(int, optional): Sleep time between download,
                unit is second.
            download_filename_format: File download local save style,
                "true": tape task ID and scene name,
                "false" : download directly without doing processing.
            local_path (str): Download file locally save path,
                default Window system is "USERPROFILE" environment variable address splicing "renderfarm_sdk",
                Linux system is "HOME" environment variable address splicing "renderfarm_sdk".
            engine_type (str, optional): set engine type, support "aspera" and "raysync", Default "aspera".
            server_ip (str, optional): transmit server host,
                if not set, it is obtained from the default transport profile.
            server_port (str, optional): transmit server port,
                if not set, it is obtained from the default transport profile.

        Returns:
            bool: True is success.

        """
        local_path = self._check_local_path(local_path)
        self.logger.info("[Rayvision_sync start"
                         "auto_download_after_task_completed .....]")
        self._download_log(task_id_list, local_path)

        while True:
            if task_id_list:
                time.sleep(float(sleep_time))
                for task_id in task_id_list:
                    is_task_end = self.manage_task.is_task_end(task_id)

                    if is_task_end is True:
                        time.sleep(float(5))
                        self.logger.info('The tasks end: %s', task_id)
                        self._run_download([task_id], local_path,
                                           max_speed, print_log,
                                           download_filename_format,
                                           engine_type=engine_type, server_ip=server_ip, server_port=server_port
                                           )
                        task_id_list.remove(task_id)
            else:
                break
        self.logger.info("[Rayvision_sync end -- "
                         "auto_download_after_task_completed......]")

        return True

    def _run_download(self, task_id_list, local_path, max_speed=None,
                      print_log=True, download_filename_format="true",
                      server_path=None, engine_type="aspera", server_ip=None, server_port=None):
        """Execute the cmd command for multitasking download.

        Args:
            task_id_list (list of int): Task id list.
            local_path (str): Download to local path.
            max_speed (str): Maximum transmission speed, default value
                is 1048576 KB/S.
            print_log (bool): Print log, True: print, False: not print.
            download_filename_format: File download local save style,
                "true": tape task ID and scene name,
                "false" : download directly without doing processing.
            server_path (str / list): The user customizes the file structure to be downloaded from
                the output server, and all file structures are downloaded by default,
                example: "18164087_test/l_layer".
            engine_type (str, optional): set engine type, support "aspera" and "raysync", Default "aspera".
            server_ip (str, optional): transmit server host,
                if not set, it is obtained from the default transport profile.
            server_port (str, optional): transmit server port,
                if not set, it is obtained from the default transport profile.

        """
        transmit_type = 'download_path'
        local_path = str2unicode(local_path)
        # The unit of 'max_speed' is KB/S, default value is 1048576 KB/S,
        #  means 1 GB/S.
        max_speed = max_speed if max_speed is not None else "1048576"

        if not server_path:
            task_status_list = self.manage_task.get_task_status(task_id_list)
            output_file_names = (self.manage_task.
                                 output_file_names(task_status_list))
        else:
            if isinstance(server_path, str):
                output_file_names = [server_path]
            elif isinstance(server_path, list):
                isinstance(server_path, list)
                output_file_names = server_path
            else:
                raise Exception("custom_server_output_path must a list or str.")

        for output_file_name in output_file_names:
            cmd_params = [transmit_type, local_path, output_file_name,
                          max_speed, download_filename_format, 'output_bid']

            cmd = self.trans.create_cmd(cmd_params, engine_type=engine_type,
                                        server_ip=server_ip, server_port=server_port)
            run_cmd(cmd, print_log=print_log, logger=self.logger)



"""Test the rayvision_sync download functions."""

import os

# pylint: disable=import-error
import pytest
from rayvision_sync.exception import RayvisionError


def test_upload_config(rayvision_upload, tmpdir):
    """Test upload_config, we can get a expected result."""
    task_id = 10101
    config_file_list = [
        str(tmpdir.join('task.json')),
        str(tmpdir.join('asset.json')),
        str(tmpdir.join('tips.json')),
        str(tmpdir)
    ]
    with pytest.raises(RayvisionError):
        rayvision_upload.upload_config(task_id, config_file_list)


def test_create_db_ini(rayvision_upload, tmpdir):
    """Test create_db_ini, we can get a db file."""
    upload_json_path = str(tmpdir.join("upload.json"))
    rayvision_upload.create_db_ini(upload_json_path)
    assert len(os.listdir(rayvision_upload.db_ini)) > 0
